import asyncio
import logging

import asyncio_atexit
import backoff
import gql
import websockets
from gql.transport.aiohttp import AIOHTTPTransport
from gql.transport.exceptions import TransportQueryError

from tibber import API_ENDPOINT
from tibber.exceptions import APIException, UnauthenticatedException

_logger = logging.getLogger(__name__)


class QueryExecutor:
    """A class for executing queries."""

    def __init__(self, session=None, transport_kwargs={}):
        self.gql_client = None
        transport = AIOHTTPTransport(
            url=API_ENDPOINT,
            headers={"Authorization": "Bearer " + self.token},
            ssl=True,
            **transport_kwargs,
        )
        self.gql_client = gql.Client(
            transport=transport, fetch_schema_from_transport=True
        )

        asyncio.run(self.__ainit__(session))

    async def __ainit__(self, session):
        self.session = session or await self.gql_client.connect_async()
        asyncio_atexit.register(self.gql_client.close_async)

    def execute_query(
        self, access_token: str, query: str, max_tries: int = 1, **kwargs
    ):
        """Executes a GraphQL query to the Tibber API.

        :param access_token: The Tibber API token to use for the request.
        :param query: The query to send to the Tibber API.
        :param max_tries: The amount of attempts before giving up. Set to None for infinite tries.
        :param **kwargs: Arguments to be passed in to the backoff.on_exception decorator
        """
        # To allow invocations from async contexts, check if a loop is running and attempt
        # to schedule the execute_async method there. If no loop is found or the loop is not
        # running, asyncio.run can be run instead.
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None

        if loop and loop.is_running():
            return loop.run_until_complete(
                self.execute_async(access_token, query, max_tries, **kwargs)
            )

        return asyncio.run(self.execute_async(access_token, query, max_tries, **kwargs))

    async def execute_async(
        self, access_token: str, query: str, max_tries: int = 1, **kwargs
    ):
        """Coroutine for executing a GraphQL query to the Tibber API asynchronously.

        :param access_token: The Tibber API token to use for the request.
        :param query: The query to send to the Tibber API.
        :param max_tries: The amount of attempts before giving up. Set to None for infinite tries.
        :param **kwargs: Arguments to be passed in to the backoff.on_exception decorator
        """
        backoff_execution = backoff.on_exception(
            backoff.expo,
            [
                gql.transport.exceptions.TransportClosed,
                websockets.exceptions.ConnectionClosedError,
            ],
            max_tries=max_tries,
            max_time=100,
            jitter=backoff.full_jitter,
            on_success=self._success_handler,
            on_backoff=self._backoff_handler,
            on_giveup=self._giveup_handler,
            **kwargs,
        )(self.execute_async_single)

        result = await backoff_execution(access_token, query)
        return result

    async def execute_async_single(self, access_token: str, query: str):
        try:
            result = await self.gql_client.execute_async(gql.gql(query))
        except TransportQueryError as e:
            for error in e.errors:
                self._process_error(error)
        except asyncio.exceptions.TimeoutError:
            _logger.error(
                "Timed out when executing a query. Check your connection to the Tibber API or the Tibber API status."
            )
            _logger.debug("query information:\n" + query)
            raise APIException("Timed out when executing query.")
        return result

    def _process_error(self, error):
        try:
            code = error["extensions"]["code"]
            message = error["message"]
        except KeyError:
            raise APIException(error)

        if code == "UNAUTHENTICATED":
            raise UnauthenticatedException(message)

        raise APIException(error)

    def _success_handler(self, details): ...

    def _backoff_handler(self, details):
        _logger.warning(
            "Backing off after {tries} tries. Calling {target} in {wait:.1f} seconds.".format(
                **details
            )
        )

    def _giveup_handler(self, details):
        _logger.error(
            "Gave up running {target} after {tries} tries. {elapsed:.1f} seconds have passed.".format(
                **details
            )
        )



from __future__ import annotations

"""A class representing the SubscriptionPriceConnectionPageInfo type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from tibber.account import Account


class SubscriptionPriceConnectionPageInfo:
    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def end_cursor(self) -> str:
        return self.cache.get("endCursor")

    @property
    def has_next_page(self) -> bool:
        return self.cache.get("hasNextPage")

    @property
    def has_previous_page(self) -> bool:
        return self.cache.get("hasPreviousPage")

    @property
    def start_cursor(self) -> str:
        return self.cache.get("startCursor")

    @property
    def resolution(self) -> str:
        return self.cache.get("resolution")

    @property
    def currency(self) -> str:
        return self.cache.get("currency")

    @property
    def count(self) -> int:
        return self.cache.get("count")

    @property
    def precision(self) -> int:
        return self.cache.get("precision")

    @property
    def min_energy(self) -> float:
        return self.cache.get("minEnergy")

    @property
    def min_total(self) -> float:
        return self.cache.get("minTotal")

    @property
    def max_energy(self) -> float:
        return self.cache.get("maxEnergy")

    @property
    def max_total(self) -> float:
        return self.cache.get("maxTotal")



from __future__ import annotations

"""A class representing the LegalEntity type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

from tibber.types.address import Address
from tibber.types.contact_info import ContactInfo

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account


class LegalEntity:
    """A LegalEntity (most commonly an owner of a home for example). This class contains methods to get information such as address, name and contact info."""

    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def id(self) -> str:
        return self.cache.get("id")

    @property
    def first_name(self) -> str:
        """First/Christian name of the entity"""
        return self.cache.get("firstName")

    @property
    def is_company(self) -> bool:
        """'true' if the entity is a company"""
        return self.cache.get("isCompany")  # pragma: no cover

    @property
    def name(self) -> str:
        """Full name of the entity"""
        return self.cache.get("name")

    @property
    def middle_name(self) -> str:
        """Middle name of the entity"""
        return self.cache.get("middleName")  # pragma: no cover

    @property
    def last_name(self) -> str:
        """Last name of the entity"""
        return self.cache.get("lastName")

    @property
    def organization_no(self) -> str:
        """Organization number - only populated if entity is a company (isCompany=true)"""
        return self.cache.get("organizationNo")  # pragma: no cover

    @property
    def language(self) -> str:
        """The primary language of the entity"""
        return self.cache.get("language")  # pragma: no cover

    @property
    def contact_info(self) -> ContactInfo:
        """Contact information of the entity"""
        return ContactInfo(self.cache.get("contactInfo"), self.tibber_client)

    @property
    def address(self) -> Address:
        """Address information for the entity"""
        return Address(
            self.cache.get("address"), self.tibber_client
        )  # pragma: no cover



from __future__ import annotations

"""A class representing the PriceRatingEntry type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account


class PriceRatingEntry:
    """A class to get the rating of a price in relative terms."""

    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def time(self) -> str:
        """The start time of the price"""
        return self.cache.get("time")

    @property
    def energy(self) -> float:
        """Nordpool spot price"""
        return self.cache.get("energy")

    @property
    def total(self) -> float:
        """The total price (incl. tax)"""
        return self.cache.get("total")

    @property
    def tax(self) -> float:
        """The tax part of the price (guarantee of origin certificate, energy tax (Sweden only) and VAT)"""
        return self.cache.get("tax")

    @property
    def difference(self) -> float:
        """The percentage difference compared to the trailing price average (1 day for 'hourly', 30 days for 'daily' and 32 months for 'monthly')"""
        return self.cache.get("difference")

    @property
    def level(self) -> str:
        """The price level compared to recent price values (calculated using 'difference' and 'priceRating.thresholdPercentages')"""
        return self.cache.get("level")



from __future__ import annotations

"""A class representing the HomeConsumptionPageInfo type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account


class HomeConsumptionPageInfo:
    """A class containing household electricity consumption information for a time period."""

    def __init__(self, resolution: str, data: dict, tibber_client: "Account"):
        self.resolution = resolution
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def end_cursor(self) -> str:
        """The global ID of the last element in the list"""
        return self.cache.get("endCursor")

    @property
    def has_next_page(self) -> bool:
        """True if further pages are available"""
        return self.cache.get("hasNextPage")

    @property
    def has_previous_page(self) -> bool:
        """True if previous pages are available"""
        return self.cache.get("hasPreviousPage")

    @property
    def start_cursor(self) -> str:
        """The global ID of the first element in the list"""
        return self.cache.get("startCursor")

    @property
    def count(self) -> int:
        """The number of elements in the list"""
        return self.cache.get("count")

    @property
    def currency(self) -> str:
        """The currency of the page"""
        return self.cache.get("currency")

    @property
    def total_cost(self) -> float:
        """Page total cost"""
        return self.cache.get("totalCost")

    @property
    def total_consumption(self) -> float:
        """Total consumption for page"""
        return self.cache.get("totalConsumption")

    @property
    def filtered(self) -> int:
        """Number of entries that have been filtered from result set due to empty nodes"""
        return self.cache.get("filtered")



from __future__ import annotations

"""A class representing the Price type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account


class Price:
    """A class to get price info."""

    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def total(self) -> float:
        """The total price (energy + taxes)"""
        return self.cache.get("total")

    @property
    def energy(self) -> float:
        """Nordpool spot price"""
        return self.cache.get("energy")

    @property
    def tax(self) -> float:
        """The tax part of the price (guarantee of origin certificate, energy tax (Sweden only) and VAT)"""
        return self.cache.get("tax")

    @property
    def starts_at(self) -> str:
        """The start time of the price"""
        return self.cache.get("startsAt")

    @property
    def currency(self) -> str:
        """The price currency"""
        return self.cache.get("currency")

    @property
    def level(self) -> str:
        """The price level compared to recent price values"""
        return self.cache.get("level")



from __future__ import annotations

"""A class representing the HomeProductionConnection type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

from tibber.types.home_production_edge import HomeProductionEdge
from tibber.types.home_production_page_info import HomeProductionPageInfo
from tibber.types.production import Production

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account


class HomeProductionConnection:
    """A class containing household electricity production information for a time period."""

    def __init__(self, resolution: str, data: dict, tibber_client: "Account"):
        self.resolution = resolution
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def page_info(self) -> str:
        return HomeProductionPageInfo(
            self.resolution, self.cache.get("pageInfo"), self.tibber_client
        )

    @property
    def nodes(self) -> list:
        return [
            Production(node, self.tibber_client) for node in self.cache.get("nodes")
        ]

    @property
    def edges(self) -> list:
        return [
            HomeProductionEdge(self.resolution, edge, self.tibber_client)
            for edge in self.cache.get("edges")
        ]

    def __iter__(self):
        return iter(self.nodes)



from __future__ import annotations

"""A class representing the SubscriptionPriceEdge type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

from tibber.types.price import Price

if TYPE_CHECKING:
    from tibber.account import Account


class SubscriptionPriceEdge:
    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def cursor(self) -> str:
        return self.cache.get("cursor")

    @property
    def node(self) -> Price:
        return Price(self.cache.get("node"), self.tibber_client)



from __future__ import annotations

"""A class representing the MeteringPointData type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account


class MeteringPointData:
    """A MeteringPointData type to get information about the grid company, metering point ID and other information of a TibberHome."""

    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def consumption_ean(self) -> str:
        """The metering point ID of the home"""
        return self.cache.get("consumptionEan")

    @property
    def grid_company(self) -> str:
        """The grid provider of the home"""
        return self.cache.get("gridCompany")

    @property
    def grid_area_code(self) -> str:
        """The grid area the home/metering point belongs to"""
        return self.cache.get("gridAreaCode")

    @property
    def price_area_code(self) -> str:
        """The price area the home/metering point belongs to"""
        return self.cache.get("priceAreaCode")

    @property
    def production_ean(self) -> str:
        """The metering point ID of the production"""
        return self.cache.get("productionEan")

    @property
    def energy_tax_type(self) -> str:
        """The eltax type of the home (only relevant for Swedish homes)"""
        return self.cache.get("energyTaxType")

    @property
    def vat_type(self) -> str:
        """The VAT type of the home (only relevant for Norwegian homes)"""
        return self.cache.get("vatType")

    @property
    def estimated_annual_consumption(self) -> int:  # pragma: no cover
        """The estimated annual consumption as reported by grid company"""
        return self.cache.get("estimatedAnnualConsumption")



from __future__ import annotations

from tibber.types.address import Address
from tibber.types.contact_info import ContactInfo
from tibber.types.home import NonDecoratedTibberHome, TibberHome
from tibber.types.home_features import HomeFeatures
from tibber.types.legal_entity import LegalEntity
from tibber.types.metering_point_data import MeteringPointData
from tibber.types.price import Price
from tibber.types.price_info import PriceInfo
from tibber.types.price_rating import PriceRating
from tibber.types.price_rating_entry import PriceRatingEntry
from tibber.types.price_rating_type import PriceRatingType
from tibber.types.subscription import Subscription
from tibber.types.viewer import Viewer

from tibber.types.price_rating_threshold_percentages import (  # isort:skip
    PriceRatingThresholdPercentages,
)

__all__ = [
    "Address",
    "ContactInfo",
    "NonDecoratedTibberHome",
    "TibberHome",
    "HomeFeatures",
    "LegalEntity",
    "MeteringPointData",
    "Price",
    "PriceInfo",
    "PriceRating",
    "PriceRatingEntry",
    "PriceRatingThresholdPercentages",
    "PriceRatingType",
    "Subscription",
    "Viewer",
]



__version__ = "0.7.1"
DEMO_TOKEN = "3A77EECF61BD445F47241A5A36202185C35AF3AF58609E19B53F3A8872AD7BE1-1"
API_ENDPOINT = "https://api.tibber.com/v1-beta/gql"

import asyncio
import os

# The event loop type causes problems on windows systems when exiting.
if os.name == "nt":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

# Import modules after defining constants to avoid circular import error.
from .account import Account
from .types.home import NonDecoratedTibberHome, TibberHome

__all__ = [
    "__version__",
    "DEMO_TOKEN",
    "API_ENDPOINT",
    "Account",
    "NonDecoratedTibberHome",
    "TibberHome",
]



from __future__ import annotations

"""A class representing the SubscriptionPriceConnection type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

from tibber.types.price import Price
from tibber.types.subscription_price_edge import SubscriptionPriceEdge

from tibber.types.subscription_price_connection_page_info import (  # isort:skip
    SubscriptionPriceConnectionPageInfo,
)

if TYPE_CHECKING:
    from tibber.account import Account


class SubscriptionPriceConnection:
    """A class to get subscription price connection."""

    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def edges(self) -> list[SubscriptionPriceEdge]:
        return [
            SubscriptionPriceEdge(edge, self.tibber_client)
            for edge in self.cache.get("edges", [])
        ]

    @property
    def page_info(self) -> SubscriptionPriceConnectionPageInfo:
        return SubscriptionPriceConnectionPageInfo(
            self.cache.get("pageInfo"), self.tibber_client
        )

    @property
    def nodes(self) -> list[Price]:
        """List of Price objects from the executed range query."""
        return [Price(node, self.tibber_client) for node in self.cache.get("nodes", [])]



from __future__ import annotations

"""A class representing the ContactInfo type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account


class ContactInfo:
    """A class to get email and mobile. This info is probably of a LegalEntity."""

    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def email(self) -> str:
        """The email of the corresponding entity"""
        return self.cache.get("email")

    @property
    def mobile(self) -> str:
        """The mobile phone no of the corresponding entity"""
        return self.cache.get("mobile")  # pragma: no cover



import json
import logging

from .networking import QueryBuilder, QueryExecutor
from .types.push_notification_response import PushNotificationResponse
from .types.viewer import Viewer

_logger = logging.getLogger(__name__)


class Account(QueryExecutor):
    """The main Tibber class to communicate with the Tibber API."""

    def __init__(
        self,
        token: str,
        user_agent: str = None,
        immediate_update: bool = True,
        transport_kwargs={},
    ):
        """Initialize the tibber client.

        :param token: The token to log in with
        :param immediate_update: Specifies whether to immediately update all tibber information
            on initialization.
        :throws UnauthenticatedException: If the provided token was not accepted by the Tibber API.
            Note that this will only be checked if the immediate_update parameter is set to True.
        """
        self.cache: dict = {}
        self._token: str = token
        self.user_agent = user_agent

        super().__init__(transport_kwargs=transport_kwargs)

        if immediate_update:
            self.fetch_all()

    def update(self):
        """Alias for fetch_all()"""
        self.fetch_all()

    async def update_async(self, retries=1):
        """Fetches all available data from the API and caches it. This method is used in async
        contexts to avoid errors about an event loop already running."""
        data = await self.execute_async(
            self.token, QueryBuilder.query_all_data(), retries
        )
        self.update_cache(data)

    def fetch_all(self, retries=1):
        """Fetches all available data from the API and caches it."""
        data = self.execute_query(self.token, QueryBuilder.query_all_data())
        self.update_cache(data)

    def update_cache(self, data):
        """Updates the cache with values from data.

        :param data: The data to add / update values in the cache with.
        """
        _logger.debug("Overwriting the cache data.")
        _logger.debug("Old data: " + json.dumps(self.cache))
        _logger.debug("New data: " + json.dumps(data))
        self.cache = QueryBuilder.combine_dicts(self.cache, data)

    def send_push_notification(
        self, title: str, message: str, screen_to_open: str = None
    ):
        """Sends a push notification to all registered devices connected to the account owning the API key."""
        data = QueryBuilder.send_push_notification(title, message, screen_to_open)
        response_data = self.execute_query(self.token, data).get("sendPushNotification")
        return PushNotificationResponse(response_data, self)

    @property
    def token(self) -> str:
        return self._token

    @token.setter
    def token(self, token: str):
        if not isinstance(token, str):
            _logger.error(
                "Attempted to set the token to a non-string datatype: " + type(token)
            )
            raise TypeError("The token must be a string.")
        self._token = token
        _logger.debug("The tibber token was set to a new value.")

    @property
    def viewer(self):
        return Viewer(self.cache.get("viewer"), self)

    @property
    def name(self):
        return self.viewer.name

    @property
    def login(self):
        return self.viewer.login

    @property
    def user_id(self):
        """Unique user identifier"""
        return self.viewer.user_id

    @property
    def account_type(self):
        """The type of account for the logged-in user."""
        return self.viewer.account_type

    @property
    def homes(self):
        """All homes visible to the logged-in user"""
        return self.viewer.homes



from __future__ import annotations

"""A class representing the PriceRatingType type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING, List

from tibber.types.price_rating_entry import PriceRatingEntry

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account


class PriceRatingType:
    """A class to get the rating of a price in relative terms."""

    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def min_energy(self) -> float:
        """Lowest Nordpool spot price over the time period"""
        return self.cache.get("minEnergy")

    @property
    def max_energy(self) -> float:
        """Highest Nordpool spot price over the time period"""
        return self.cache.get("maxEnergy")

    @property
    def min_total(self) -> float:
        """Lowest total price (incl. tax) over the time period"""
        return self.cache.get("minTotal")

    @property
    def max_total(self) -> float:
        """Highest total price (incl. tax) over the time period"""
        return self.cache.get("maxTotal")

    @property
    def currency(self) -> str:
        """The price currency"""
        return self.cache.get("currency")

    @property
    def entries(self) -> List[PriceRatingEntry]:
        """The individual price entries aggregated by hourly/daily/monthly values"""
        return [
            PriceRatingEntry(entry, self.tibber_client)
            for entry in self.cache.get("entries", [])
        ]



from setuptools import setup

from tibber import __version__

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="tibber.py",
    long_description=long_description,
    long_description_content_type='text/markdown',
    packages=["tibber", "tibber.networking", "tibber.exceptions", "tibber.types"],
    install_requires=[
        "gql==3.5.3",
        "gql[aiohttp]>=3.4.0",
        "gql[websockets]>=3.4.0",
        "graphql-core>=3.2.3",
        "backoff>=2.2.1",
        "asyncio-atexit>=1.0.1",
    ],
    license="MIT",
    version=__version__,
    description="A python wrapper for the Tibber API.",
    python_requires=">=3.9.0",
    author="BeatsuDev",
    author_email="",
    url="https://tibberpy.readthedocs.io/en/latest/",
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Documentation :: Sphinx",
        "Topic :: Home Automation",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
)



from __future__ import annotations

"""A class representing the PriceInfo type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING, Optional

from tibber.networking.query_builder import QueryBuilder
from tibber.types.price import Price

from tibber.types.subscription_price_connection import (  # isort: skip
    SubscriptionPriceConnection,
)

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account


class PriceInfo:
    """A class to get price info."""

    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def current(self) -> Price:
        """The energy price right now"""
        return Price(self.cache.get("current"), self.tibber_client)

    @property
    def today(self) -> list[Price]:
        """The hourly prices of the current day"""
        return [Price(hour, self.tibber_client) for hour in self.cache.get("today", [])]

    @property
    def tomorrow(self) -> list[Price]:
        """The hourly prices of the upcoming day"""
        return [
            Price(hour, self.tibber_client) for hour in self.cache.get("tomorrow", [])
        ]

    def fetch_range(
        self,
        resolution: str,
        first: Optional[str] = None,
        last: Optional[str] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
        home_id: Optional[str] = None,
    ) -> SubscriptionPriceConnection:
        """Fetch the price range.

        The before and after arguments are Base64 encoded ISO 8601 datetimes."""
        range_query_dict = QueryBuilder.range_query(
            resolution, first, last, before, after
        )

        range_query = QueryBuilder.create_query(
            "viewer", "homes", "currentSubscription", "priceInfo", range_query_dict
        )
        full_data = self.tibber_client.execute_query(
            self.tibber_client.token, range_query
        )

        home = full_data["viewer"]["homes"][0]
        if home_id:
            home_of_id = [
                home for home in full_data["viewer"]["homes"] if home["id"] == home_id
            ][0]

            if home_of_id:
                home = home_of_id

        return SubscriptionPriceConnection(
            home["currentSubscription"]["priceInfo"]["range"], self.tibber_client
        )



from __future__ import annotations

"""A class representing the Subscription type from the GraphQL Tibber API."""
from typing import TYPE_CHECKING, Literal, Optional

from tibber.networking.query_builder import QueryBuilder
from tibber.types.legal_entity import LegalEntity
from tibber.types.price_info import PriceInfo
from tibber.types.price_rating import PriceRating
from tibber.types.subscription_price_connection import SubscriptionPriceConnection

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account


class Subscription:
    """A class to get information about the subscription of a TibberHome."""

    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    @property
    def id(self) -> str:
        return self.cache.get("id")

    @property
    def subscriber(self) -> LegalEntity:
        """The owner of the subscription"""
        return LegalEntity(self.cache.get("subscriber"), self.tibber_client)

    @property
    def valid_from(self) -> str:
        """The time the subscription started"""
        return self.cache.get("validFrom")

    @property
    def valid_to(self) -> str:
        """The time the subscription ended"""
        return self.cache.get("validTo")

    @property
    def status(self) -> str:
        """The current status of the subscription"""
        return self.cache.get("status")

    @property
    def price_info(self) -> PriceInfo:
        """Price information related to the subscription"""
        return PriceInfo(self.cache.get("priceInfo"), self.tibber_client)

    def fetch_price_info(
        self,
        resolution: Literal["HOURLY", "QUARTER_HOURLY"],
        home_id: Optional[str] = None,
    ) -> PriceInfo:
        price_info_query_dict = QueryBuilder.price_info_query(resolution)

        price_info_query = QueryBuilder.create_query(
            "viewer", "homes", "currentSubscription", price_info_query_dict
        )

        full_data = self.tibber_client.execute_query(
            self.tibber_client.token, price_info_query
        )

        home = full_data["viewer"]["homes"][0]
        if home_id:
            home_of_id = [
                home for home in full_data["viewer"]["homes"] if home["id"] == home_id
            ][0]

            if home_of_id:
                home = home_of_id

        return PriceInfo(home["currentSubscription"]["priceInfo"], self.tibber_client)

    def fetch_price_info_range(
        self,
        resolution: str,
        first: Optional[str] = None,
        last: Optional[str] = None,
        before: Optional[str] = None,
        after: Optional[str] = None,
        home_id: Optional[str] = None,
    ) -> SubscriptionPriceConnection:
        """Fetch PriceInfo for a given range.

        The before and after arguments are Base64 encoded ISO 8601 datetimes."""
        price_info_range_query_dict = QueryBuilder.price_info_range_query(
            resolution, first, last, before, after
        )

        price_info_range_query = QueryBuilder.create_query(
            "viewer", "homes", "currentSubscription", price_info_range_query_dict
        )
        full_data = self.tibber_client.execute_query(
            self.tibber_client.token, price_info_range_query
        )

        home = full_data["viewer"]["homes"][0]
        if home_id:
            home_of_id = [
                home for home in full_data["viewer"]["homes"] if home["id"] == home_id
            ][0]

            if home_of_id:
                home = home_of_id

        return SubscriptionPriceConnection(
            home["currentSubscription"]["priceInfoRange"], self.tibber_client
        )

    @property
    def price_rating(self) -> PriceRating:
        """Price information related to the subscription"""
        return PriceRating(self.cache.get("priceRating"), self.tibber_client)



"""A class for generating GraphQL queries to send to the Tibber API"""

from typing import Literal


class QueryBuilder:
    @classmethod
    def create_query_from_dict(
        cls, query_dict: dict, indentation: int = 0, last: bool = True
    ) -> str:
        """Creates a graphQL query from the keys of a dictionary.

        :param query_dict: The dictionary to convert to a query.
        :param indentation: The amount of spaces to indent the returned query.
        :param last: Specifies if the function call is the last in a recursive loop.
        """
        string_query = "{\n"

        for key, value in query_dict.items():
            if isinstance(value, dict):
                string_query += " " * (indentation + 2) + key + " "
                string_query += cls.create_query_from_dict(
                    value, indentation + 2, False
                )
                string_query += " " * (indentation + 2) + "}\n"
            else:
                string_query += " " * (indentation + 2) + key + "\n"

        if last:
            string_query += "}"

        return string_query

    @classmethod
    def create_query(cls, *args):
        """Creates a query given keys. The last argument must be a valid property of QueryBuilder that fits
        according to the GraphQL structure of the Tibber API.

        Exmaple:
            create_query("viewer", "homes", "currentSubscription", QueryBuilder.price_info())

            This returns a string query that queries specifically for the
            price_info of the current subscription in all homes.
        """
        if len(args) == 0:
            raise TypeError(
                "The QueryBuilder.create_query method requires at least one argument!"
            )

        def nest_dict(*keys):
            if len(keys) == 1:
                return keys[0]
            return {keys[0]: nest_dict(*keys[1:])}

        return QueryBuilder.create_query_from_dict(nest_dict(*args))

    @classmethod
    def combine_dicts(cls, dict1: dict, dict2: dict) -> dict:
        """Combines two nested dictionaries. The values from second dictionary overwrites the first one
        if the keys are exactly the same. The only exception to this is if the first dictionary's value
        is a dict, while the second dictionary value is a string. Then the dictionary with a dictionary
        typed value is prioritized.

        This method is meant to be used to combine query building components in dictionary form.

        :param dict1: The first dict to combine with the second dict
        :param dict2: The second dict to combine with the first dict
        :throws TypeError: If none of the parameters are dicts
        """
        if not (isinstance(dict1, dict) and isinstance(dict2, dict)):
            raise TypeError(f"Cannot combine types {type(dict1)} and {type(dict2)}.")

        result_dict = dict1

        for key, value in dict2.items():
            # The key in the second dict does not exist in the first so it's safe to add it
            # without overwriting the value from dict1 (because there is none)
            if key not in dict1.keys():
                result_dict[key] = value

            # If the value in dict1 is a string, and we know the key exists in both dict1
            # and dict2, the value from dict2 overwrite the value in dict1.
            elif isinstance(dict1[key], str):
                result_dict[key] = value

            # We know the key already exists in both dicts now. If the values in dict1 and dict 2
            # are also dicts, we need to call the function again on these dicts to resolve "merge conflicts"
            elif isinstance(value, dict):
                result_dict[key] = cls.combine_dicts(dict1[key], dict2[key])

            # We know the key exists in both dicts. If the value in dict1 is a dictionary, but
            # the value in dict2 is a string, the dict1 value is prioritized.
            elif isinstance(dict1[key], dict):
                # We can continue because the values of dict1 is already in the resulting dict
                continue

            # If the value in both dicts are strings.
            else:
                result_dict[key] = value

        return result_dict

    @classmethod
    def query_all_data(cls) -> str:
        return cls.create_query_from_dict(QueryBuilder.query())

    # -------------------------------------------------------------------
    # Query dicts for the Tibber API types. Remember that values ignored.
    # -------------------------------------------------------------------

    @classmethod
    def query(cls) -> dict:
        """Return a dict with query values as keys for all information on the `Query` type. This type is the base type
        which all queries are nested in. Therefore, this query returns all information available from the api.
        """
        return {"viewer": QueryBuilder.viewer()}

    @classmethod
    def viewer(cls) -> dict:
        """Return a dict with query values as keys for all information on the `Viewer` type."""
        return {
            "login": "",
            "userId": "",
            "name": "",
            "accountType": [],
            "homes": QueryBuilder.home(),
            "websocketSubscriptionUrl": "",
        }

    @classmethod
    def home(cls) -> dict:
        """Return a dict with query values as keys for all information on the `Home` type."""
        return {
            "id": "",
            "timeZone": "",
            "appNickname": "",
            "appAvatar": "",
            "size": 0,
            "type": "",
            "numberOfResidents": 0,
            "primaryHeatingSource": "",
            "hasVentilationSystem": False,
            "mainFuseSize": 0,
            "address": QueryBuilder.address(),
            "owner": QueryBuilder.legal_entity(),
            "meteringPointData": QueryBuilder.metering_point_data(),
            "currentSubscription": QueryBuilder.subscription(),
            "subscriptions": QueryBuilder.subscription(),
            "features": QueryBuilder.features(),
        }

    @classmethod
    def address(cls) -> dict:
        """Return a dict with query values as keys for all information on the `Address` type."""
        return {
            "address1": "",
            "address2": "",
            "address3": "",
            "city": "",
            "postalCode": "",
            "country": "",
            "latitude": "",
            "longitude": "",
        }

    @classmethod
    def legal_entity(cls) -> dict:
        """Return a dict with query values as keys for all information on the `LegalEntity` type."""
        return {
            "id": "",
            "firstName": "",
            "isCompany": False,
            "name": "",
            "middleName": "",
            "lastName": "",
            "organizationNo": "",
            "language": "",
            "contactInfo": QueryBuilder.contact_info(),
            "address": QueryBuilder.address(),
        }

    @classmethod
    def metering_point_data(cls) -> dict:
        """Return a dict with query values as keys for all information on the `MeteringPointData` type."""
        return {
            "consumptionEan": "",
            "gridCompany": "",
            "gridAreaCode": "",
            "priceAreaCode": "",
            "productionEan": "",
            "energyTaxType": "",
            "vatType": "",
            "estimatedAnnualConsumption": 0,
        }

    @classmethod
    def subscription(cls) -> dict:
        """Return a dict with query values as keys for all information on the `Subscription` type."""
        return {
            "id": "",
            "subscriber": QueryBuilder.legal_entity(),
            "validFrom": "",
            "validTo": "",
            "status": "",
            "priceInfo": QueryBuilder.price_info(),
            "priceRating": QueryBuilder.price_rating(),
        }

    @classmethod
    def features(cls) -> dict:
        """Return a dict with query values as keys for all information on the `HomeFeatures` type."""
        return {"realTimeConsumptionEnabled": False}

    @classmethod
    def contact_info(cls) -> dict:
        """Return a dict with query values as keys for all information on the `ContactInfo` type."""
        return {"email": "", "mobile": ""}

    @classmethod
    def price_info(cls) -> dict:
        """Return a dict with query values as keys for all information on the `PriceInfo` type."""
        return {
            "current": QueryBuilder.price(),
            "today": QueryBuilder.price(),
            "tomorrow": QueryBuilder.price(),
        }

    @classmethod
    def price_rating(cls) -> dict:
        """Return a dict with query values as keys for all information on the `PriceRating` type."""
        return {
            "thresholdPercentages": QueryBuilder.price_rating_threshold_percentages(),
            "hourly": QueryBuilder.price_rating_type(),
            "daily": QueryBuilder.price_rating_type(),
            "monthly": QueryBuilder.price_rating_type(),
        }

    @classmethod
    def price(cls) -> dict:
        """Return a dict with query values as keys for all information on the `Price` type."""
        return {
            "total": 0.0,
            "energy": 0.0,
            "tax": 0.0,
            "startsAt": "",
            "currency": "",
            "level": "",
        }

    @classmethod
    def price_rating_threshold_percentages(cls) -> dict:
        """Return a dict with query values as keys for all information on the `PriceRatingThresholdPercentages` type."""
        return {"high": 0.0, "low": 0.0}

    @classmethod
    def price_rating_type(cls) -> dict:
        """Return a dict with query values as keys for all information on the `PriceRatingType` type."""
        return {
            "minEnergy": 0.0,
            "maxEnergy": 0.0,
            "minTotal": 0.0,
            "maxTotal": 0.0,
            "currency": "",
            "entries": QueryBuilder.price_rating_entry(),
        }

    @classmethod
    def price_rating_entry(cls) -> dict:
        """Return a dict with query values as keys for all information on the `PriceRatingEntry` type."""
        return {
            "time": "",
            "energy": 0.0,
            "total": 0.0,
            "tax": 0.0,
            "difference": 0.0,
            "level": "",
        }

    # Queries with arguments. Note that these are different than queries in the way that
    # these components need to return the key for the query as well as all it's properties
    # not just the queries. This means that these should be unpacked when combined with
    # other queries, not added as a value of a key.

    @classmethod
    def single_home(home_id: str) -> dict:
        return {f"home({home_id})": QueryBuilder.home()}

    @classmethod
    def price_info_range_query(
        cls, resolution: str, first: int, last: int, before: str, after: str
    ):
        first_arg = f"first: {first}" if first else None
        last_arg = f"last: {last}" if last else None
        before_arg = f'before: "{before}"' if before else None
        after_arg = f'after: "{after}"' if after else None

        args = ", ".join(
            [
                arg
                for arg in [first_arg, last_arg, before_arg, after_arg]
                if arg is not None
            ]
        )
        return {
            f"priceInfoRange(resolution: {resolution}, {args})": {
                "pageInfo": QueryBuilder.subscription_price_connection_page_info(),
                "edges": QueryBuilder.subscription_price_edge(),
                "nodes": QueryBuilder.price(),
            }
        }

    @classmethod
    def price_info_query(cls, resolution: Literal["HOURLY", "QUARTER_HOURLY"]):
        return {
            f"priceInfo(resolution: {resolution})": {
                "current": QueryBuilder.price(),
                "today": QueryBuilder.price(),
                "tomorrow": QueryBuilder.price(),
            }
        }

    @classmethod
    def range_query(
        cls, resolution: str, first: int, last: int, before: str, after: str
    ):
        first_arg = f"first: {first}" if first else None
        last_arg = f"last: {last}" if last else None
        before_arg = f'before: "{before}"' if before else None
        after_arg = f'after: "{after}"' if after else None

        args = ", ".join(
            [
                arg
                for arg in [first_arg, last_arg, before_arg, after_arg]
                if arg is not None
            ]
        )
        return {
            f"range(resolution: {resolution}, {args})": {
                "pageInfo": QueryBuilder.subscription_price_connection_page_info(),
                "edges": QueryBuilder.subscription_price_edge(),
                "nodes": QueryBuilder.price(),
            }
        }

    @classmethod
    def consumption_query(
        cls,
        resolution: str,
        first: int = None,
        last: int = None,
        before: str = None,
        after: str = None,
        filter_empty_nodes: bool = False,
    ):
        # TODO: I feel like this can be improved...
        first_arg = f"first: {first}" if first else ""
        last_arg = f"last: {last}" if last else ""
        before_arg = f'before: "{before}"' if before else ""
        after_arg = f'after: "{after}"' if after else ""

        args = ", ".join(
            [arg for arg in [first_arg, last_arg, before_arg, after_arg] if arg]
        )
        return {
            f"consumption(resolution: {resolution}, {args}, filterEmptyNodes: {str(filter_empty_nodes).lower()})": {
                "pageInfo": QueryBuilder.home_consumption_page_info(),
                "nodes": QueryBuilder.consumption(),
                "edges": QueryBuilder.home_consumption_edge(),
            }
        }

    @classmethod
    def production_query(
        cls,
        resolution: str,
        first: int = None,
        last: int = None,
        before: str = None,
        after: str = None,
        filter_empty_nodes: bool = False,
    ):
        first_arg = f"first: {first}" if first else ""
        last_arg = f"last: {last}" if last else ""
        before_arg = f'before: "{before}"' if before else ""
        after_arg = f'after: "{after}"' if after else ""

        args = ", ".join(
            [arg for arg in [first_arg, last_arg, before_arg, after_arg] if arg]
        )
        return {
            f"production(resolution: {resolution}, {args}, filterEmptyNodes: {str(filter_empty_nodes).lower()})": {
                "pageInfo": QueryBuilder.home_production_page_info(),
                "nodes": QueryBuilder.production(),
                "edges": QueryBuilder.home_production_edge(),
            }
        }

    @classmethod
    def home_consumption_page_info(cls):
        return {
            "endCursor": "",
            "hasNextPage": False,
            "hasPreviousPage": False,
            "startCursor": "",
            "count": 0,
            "currency": "",
            "totalCost": 0.0,
            "totalConsumption": 0.0,
            "filtered": 0,
        }

    @classmethod
    def consumption(cls):
        return {
            "from": "",
            "to": "",
            "unitPrice": 0.0,
            "unitPriceVAT": 0.0,
            "consumption": 0.0,
            "consumptionUnit": "",
            "cost": 0.0,
            "currency": "",
        }

    @classmethod
    def home_consumption_edge(cls):
        return {"cursor": "", "node": QueryBuilder.consumption()}

    @classmethod
    def home_production_page_info(cls):
        return {
            "endCursor": "",
            "hasNextPage": False,
            "hasPreviousPage": False,
            "startCursor": "",
            "count": 0,
            "currency": "",
            "totalProfit": 0.0,
            "totalProduction": 0.0,
            "filtered": 0,
        }

    @classmethod
    def production(cls):
        return {
            "from": "",
            "to": "",
            "unitPrice": 0.0,
            "unitPriceVAT": 0.0,
            "production": 0.0,
            "productionUnit": "",
            "profit": 0.0,
            "currency": "",
        }

    @classmethod
    def home_production_edge(cls):
        return {"cursor": "", "node": QueryBuilder.production()}

    @classmethod
    def subscription_price_connection_page_info(cls):
        return {
            "endCursor": "",
            "hasNextPage": False,
            "hasPreviousPage": False,
            "startCursor": "",
            "resolution": "",
            "currency": "",
            "count": 0,
            "precision": "",
            "minEnergy": 0.0,
            "minTotal": 0.0,
            "maxEnergy": 0.0,
            "maxTotal": 0.0,
        }

    @classmethod
    def subscription_price_edge(cls):
        return {
            "cursor": "",
            "node": QueryBuilder.price(),
        }

    # Live data - This WILL be rewritten together with this whole class.
    # TODO: Rewrite the whole class from a dict-based approach to a string-based approach.
    @classmethod
    def live_measurement(cls, home_id):
        return f"""subscription {{
            liveMeasurement(homeId: "{home_id}") {{
                timestamp
                power
                lastMeterConsumption
                accumulatedConsumption
                accumulatedProduction
                accumulatedConsumptionLastHour
                accumulatedProductionLastHour
                accumulatedCost
                accumulatedReward
                currency
                minPower
                averagePower
                maxPower
                powerProduction
                powerReactive
                powerProductionReactive
                minPowerProduction
                maxPowerProduction
                lastMeterProduction
                powerFactor
                voltagePhase1
                voltagePhase2
                voltagePhase3
                currentL1
                currentL2
                currentL3
                signalStrength
            }}
        }}"""

    @classmethod
    def send_push_notification(
        cls, title: str, message: str, screen_to_open: str = None
    ):
        return f"""mutation {{
            sendPushNotification(input: {{
                title: "{title}",
                message: "{message}",
                {'screenToOpen: ' + screen_to_open if screen_to_open else ""}
            }}) {{
                successful
                pushedToNumberOfDevices
            }}
        }}"""



from __future__ import annotations

"""Classes representing the Home type from the GraphQL Tibber API."""
import asyncio
import inspect
import logging
from typing import TYPE_CHECKING, Callable, Union

import backoff
import gql
import websockets
from gql.transport.exceptions import TransportQueryError
from gql.transport.websockets import WebsocketsTransport
from graphql import parse

from tibber import __version__
from tibber.networking import QueryBuilder
from tibber.types.address import Address
from tibber.types.home_consumption_connection import HomeConsumptionConnection
from tibber.types.home_features import HomeFeatures
from tibber.types.home_production_connection import HomeProductionConnection
from tibber.types.legal_entity import LegalEntity
from tibber.types.live_measurement import LiveMeasurement
from tibber.types.metering_point_data import MeteringPointData
from tibber.types.subscription import Subscription

# Import type checking modules
if TYPE_CHECKING:
    from tibber.account import Account


_logger = logging.getLogger(__name__)


class NonDecoratedTibberHome:
    """A Tibber home with methods to get/fetch home information without the decorator functions to subscribe to live data."""

    def __init__(self, data: dict, tibber_client: "Account"):
        self.cache: dict = data or {}
        self.tibber_client: "Account" = tibber_client

    def fetch_consumption(
        self,
        resolution: str,
        first: int = None,
        last: int = None,
        before: str = None,
        after: str = None,
        filter_empty_nodes: bool = False,
    ) -> HomeConsumptionConnection:
        """Consumption connection"""
        consumption_query = QueryBuilder.consumption_query(
            resolution, first, last, before, after, filter_empty_nodes
        )
        full_query = QueryBuilder.create_query(
            "viewer", f'home(id: "{self.id}")', consumption_query
        )
        unsanitized_data = self.tibber_client.execute_query(
            self.tibber_client.token, full_query
        )

        # The format should be correct, since we requested it this way and the request
        # was successful, so we don't need to worry about key errors.
        data = unsanitized_data["viewer"]["home"]["consumption"]
        return HomeConsumptionConnection(resolution, data, self.tibber_client)

    def fetch_production(
        self,
        resolution: str = None,
        first: int = None,
        last: int = None,
        before: str = None,
        after: str = None,
        filter_empty_nodes: bool = False,
    ) -> HomeProductionConnection:
        production_query = QueryBuilder.production_query(
            resolution, first, last, before, after, filter_empty_nodes
        )
        full_query = QueryBuilder.create_query(
            "viewer", f'home(id: "{self.id}")', production_query
        )
        unsanitized_data = self.tibber_client.execute_query(
            self.tibber_client.token, full_query
        )

        data = unsanitized_data["viewer"]["home"]["production"]
        return HomeProductionConnection(resolution, data, self.tibber_client)

    @property
    def id(self) -> str:
        return self.cache.get("id")

    @property
    def time_zone(self) -> str:
        """The time zone the home resides in"""
        return self.cache.get("timeZone")

    @property
    def app_nickname(self) -> str:
        """The nickname given to the home by the user"""
        return self.cache.get("appNickname")

    @property
    def app_avatar(self) -> str:
        """The chosen avatar for the home"""
        return self.cache.get("appAvatar")

    @property
    def size(self) -> int:
        """The size of the home in square meters"""
        return self.cache.get("size")

    @property
    def type(self) -> str:
        """The type of home."""
        return self.cache.get("type")

    @property
    def number_of_residents(self) -> int:
        """The number of people living in the home"""
        return self.cache.get("numberOfResidents")

    @property
    def primary_heating_source(self) -> str:
        """The primary form of heating in the household"""
        return self.cache.get("primaryHeatingSource")

    @property
    def has_ventilation_system(self) -> bool:
        """Whether the home has a ventilation system"""
        return self.cache.get("hasVentilationSystem")

    @property
    def main_fuse_size(self) -> int:
        """The main fuse size"""
        return self.cache.get("mainFuseSize")

    @property
    def owner(self) -> LegalEntity:
        """The registered owner of the house"""
        return LegalEntity(self.cache.get("owner"), self.tibber_client)

    @property
    def metering_point_data(self) -> MeteringPointData:
        return MeteringPointData(
            self.cache.get("meteringPointData"), self.tibber_client
        )

    @property
    def current_subscription(self) -> Subscription:
        """The current/latest subscription related to the home"""
        return Subscription(self.cache.get("currentSubscription"), self.tibber_client)

    @property
    def subscriptions(self) -> list:
        """All historic subscriptions related to the home"""
        return [
            Subscription(sub, self.tibber_client)
            for sub in self.cache.get("subscriptions", [])
        ]

    @property
    def features(self):
        return HomeFeatures(self.cache.get("features"), self.tibber_client)

    # Support 1 to 1 Tibber API representation.
    @property
    def address(self) -> Address:
        return Address(self.cache.get("address"), self.tibber_client)

    @property
    def address1(self) -> str:
        return self.address.address1

    @property
    def address2(self) -> str:
        return self.address.address2  # pragma: no cover

    @property
    def address3(self) -> str:
        return self.address.address3  # pragma: no cover

    @property
    def city(self) -> str:
        return self.address.city  # pragma: no cover

    @property
    def postal_code(self) -> str:
        return self.address.postal_code  # pragma: no cover

    @property
    def country(self) -> str:
        return self.address.country  # pragma: no cover

    @property
    def latitude(self) -> str:
        return self.address.latitude  # pragma: no cover

    @property
    def longitude(self) -> str:
        return self.address.longitude  # pragma: no cover


class TibberHome(NonDecoratedTibberHome):
    """A Tibber home with methods to get/fetch home information and subscribe to live data.
    This class expands on the NonDecoratedTibberHome class by adding methods to subscribe to live data.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._websocket_client = None
        self._callbacks = {"live_measurement": []}

    def event(self, event_to_listen_for) -> Callable:
        """Returns a decorator that registers the function being
        decorated as a callback function for the given event

        :param event_to_listen_for: The event the decorator should register the function as a callback for.
        """

        def decorator(callback):
            """Returns the function as it is, but registers it as a callback for an event.

            :param callback: The function being decorated.
            :throws ValueError: if the given event is not a valid event.
            """
            # Check if callback is a coroutine
            if not inspect.iscoroutinefunction(callback):
                raise ValueError(
                    "Callback functions must be coroutines! Use the `async def` syntax, instead of just `def`."
                )

            # If the key is not found - the event is not a valid event!
            # Valid events will be added directly to the line where _callbacks is initialized.
            try:
                self._callbacks[event_to_listen_for].append(callback)
            except KeyError:
                raise ValueError(
                    f"Could not recognize the event you want to listen for: {event_to_listen_for}"
                )

            return callback

        return decorator

    def start_live_feed(
        self,
        user_agent=None,
        exit_condition: Callable[[LiveMeasurement], bool] = None,
        retries: int = 3,
        retry_interval: Union[float, int] = 10,
        **kwargs,
    ) -> None:
        """Creates a websocket and starts pushing data out to registered callbacks.

        :param exit_condition: A function that takes a LiveMeasurement as input and returns a boolean.
            If the function returns True, the websocket will be closed.
        :param retries: The number of times to retry connecting to the websocket if it fails.
        :param retry_interval: The interval in seconds to wait before retrying to connect to the websocket.
        :param kwargs: Additional arguments to pass to the websocket (gql.transport.WebsocketsTransport).
        """
        if not self.features.real_time_consumption_enabled:
            raise ValueError("The home does not have real time consumption enabled.")

        if not self.tibber_client.user_agent and not user_agent:
            raise ValueError(
                'You must specify a user agent when starting the live feed. E.g. "Homey/10.0.0"'
            )

        self.tibber_client.user_agent = user_agent or self.tibber_client.user_agent

        # The folllowing code is just to run the websocket loop in the correct loop.
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None

        try:
            if loop and loop.is_running():
                loop.run_until_complete(
                    self.start_websocket_loop(exit_condition, retries=retries, **kwargs)
                )
            else:
                asyncio.run(
                    self.start_websocket_loop(exit_condition, retries=retries, **kwargs)
                )
        except KeyboardInterrupt:
            _logger.info("Keyboard interrupt detected. Websocket should be closed now.")

    async def start_websocket_loop(
        self,
        exit_condition: Callable[[LiveMeasurement], bool] = None,
        retries: int = 3,
        retry_interval: Union[float, int] = 10,
        **kwargs,
    ) -> None:
        """Starts a websocket to subscribe for live measurements.

        :param exit_condition: A function that takes a LiveMeasurement as input and returns a boolean.
            If the function returns True, the websocket will be closed.
        :param retries: The number of times to retry connecting to the websocket if it fails.
        :param retry_interval: The interval in seconds to wait before retrying to connect to the websocket.
        :param kwargs: Additional arguments to pass to the websocket (gql.transport.WebsocketsTransport).
        """
        if retry_interval < 1:
            raise ValueError("The retry interval must be at least 1 second.")

        # Create the websocket
        transport = WebsocketsTransport(
            **kwargs,
            url=self.tibber_client.viewer.websocket_subscription_url,
            subprotocols=[WebsocketsTransport.GRAPHQLWS_SUBPROTOCOL],
            init_payload={"token": self.tibber_client.token},
            headers={
                "User-Agent": f"{self.tibber_client.user_agent} tibber.py/{__version__}"
            },
            ping_interval=10,
            pong_timeout=10,
        )

        self._websocket_client = gql.Client(
            transport=transport,
            fetch_schema_from_transport=True,
        )

        retry_connect = backoff.on_exception(
            backoff.expo,
            [
                gql.transport.exceptions.TransportClosed,
                websockets.exceptions.ConnectionClosedError,
            ],
            max_value=100,
            max_tries=retries,
            on_backoff=lambda details: _logger.warning(
                "Retrying to connect with backoff. Running {target} in {wait:.1f} seconds after {tries} tries.".format(
                    **details
                )
            ),
            jitter=backoff.full_jitter,
            giveup=lambda e: isinstance(e, TransportQueryError)
            or isinstance(e, ValueError),
        )

        retry_subscribe = backoff.on_exception(
            backoff.expo,
            Exception,
            max_value=100,
            max_tries=retries,
            on_backoff=lambda details: _logger.warning(
                "Retrying to subscribe with backoff. Running {target} in {wait:.1f} seconds after {tries} tries.".format(
                    **details
                )
            ),
            jitter=backoff.full_jitter,
            giveup=lambda e: isinstance(e, TransportQueryError)
            or isinstance(e, ValueError),
        )

        _logger.debug("connecting to websocket")
        session = await self._websocket_client.connect_async(
            reconnecting=True,
            retry_connect=retry_connect,
        )
        _logger.info("Connected to websocket.")
        await retry_subscribe(self.run_websocket_loop)(session, exit_condition)

    async def run_websocket_loop(self, session, exit_condition) -> None:
        # Check if real time consumption is enabled
        _logger.info(
            "Updating home information to check if real time consumption is enabled."
        )
        await self.tibber_client.update_async()

        if not self.features.real_time_consumption_enabled:
            raise ValueError("The home does not have real time consumption enabled.")

        # Subscribe to the websocket
        query = QueryBuilder.live_measurement(self.id)
        _logger.debug(
            f"Connecting to live measurement data endpoint with query: {' '.join(query.split())}"
        )
        document_node_query = parse(query)

        _logger.info("Subscribing to websocket.")
        async for data in session.subscribe(document_node_query):
            _logger.debug("real time data received.")

            # Returns True if exit condition is met
            exit_condition_met = await self.process_websocket_response(
                data, exit_condition=exit_condition
            )
            if exit_condition_met:
                _logger.info("Exit condition met. The live loop is now exiting.")
                break

        await self.close_websocket_connection()

    async def process_websocket_response(
        self, data: dict, exit_condition: Callable[[LiveMeasurement], bool] = None
    ) -> bool:
        """Processes a response with data from the live data websocket. This function will call all registered callbacks
        before checking if the exit condition is met.

        :param data: The data to process.
        :return: Returns True if exit condition was met. False otherwise.
        """
        # Broadcast the event
        # TODO: Differentiate between consumption data, production data and other data.
        cleaned_data = LiveMeasurement(data["liveMeasurement"], self.tibber_client)
        await self.broadcast_event("live_measurement", cleaned_data)

        # Check if the exit condition is met
        if exit_condition and exit_condition(cleaned_data):
            return True
        return False

    async def broadcast_event(self, event, data) -> None:
        if event not in self._callbacks:
            _logger.warning(
                f'The event "{event}" was attempted emitted, but does not exist. Nothing was run.'
            )

        if len(self._callbacks[event]) == 0:
            _logger.warning(
                "The event that was broadcasted has no listeners / callbacks! Nothing was run."
            )
            return

        await asyncio.gather(*[c(data) for c in self._callbacks[event]])

    async def close_websocket_connection(self) -> None:
        _logger.debug("attempting to close websocket connection")
        if self.websocket_running:
            try:
                await self._websocket_client.close_async()
                self._websocket_client = None  # Dereference for gc
                _logger.info("Websocket connection closed.")
            except KeyboardInterrupt as e:
                _logger.warning(
                    "Keyboard interrupt detected while closing wbsocket connection. This may cause the websocket to be left open."
                )
                raise e
        else:
            _logger.info(
                "The websocket was not running when attempting to close the websocket."
                + " The invocation of close_websocket_connection() therefore did nothing..."
            )

    @property
    def websocket_running(self) -> bool:
        """Returns True if the websocket is running. False otherwise."""
        return (
            self._websocket_client is not None
            and isinstance(self._websocket_client.transport, WebsocketsTransport)
            and self._websocket_client.transport.websocket is not None
            and self._websocket_client.transport.websocket.open
        )
